import json

import torch
from torch.utils.tensorboard import SummaryWriter
from transformers import AutoModelForQuestionAnswering, AutoTokenizer

from qaner.arg_parse import get_train_args
from qaner.data_utils import prepare_sentences_and_spans, read_bio_markup
from qaner.dataset import Collator, Dataset
from qaner.train_utils import train
from qaner.utils import set_global_seed


def main() -> int:
    """
    Main train function.

    Returns:
        int: exit code.
    """

    # argparse
    args = get_train_args()

    set_global_seed(args.seed)
    writer = SummaryWriter(log_dir=args.log_dir)
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

    # bert model
    tokenizer = AutoTokenizer.from_pretrained(args.bert_model_name)
    model = AutoModelForQuestionAnswering.from_pretrained(args.bert_model_name).to(
        device
    )

    # TODO: validate it
    tokenizer_kwargs = {
        "max_length": 512,
        "truncation": "only_second",
        "padding": True,
        "return_tensors": "pt",
        "return_offsets_mapping": True,
    }

    # dataset and dataloader (train)
    # TODO: add more general read_data function
    # TODO: add valid data

    with open(args.path_to_prompt_mapper, mode="r", encoding="utf-8") as fp:
        prompt_mapper = json.load(fp)

    train_token_seq, train_label_seq = read_bio_markup(args.path_to_train_data)

    train_qa_sentences, train_qa_labels = prepare_sentences_and_spans(
        token_seq=train_token_seq,
        label_seq=train_label_seq,
    )

    train_dataset = Dataset(
        qa_sentences=train_qa_sentences,
        qa_labels=train_qa_labels,
        prompt_mapper=prompt_mapper,
    )

    # dataset and dataloader (test)
    test_token_seq, test_label_seq = read_bio_markup(args.path_to_test_data)

    test_qa_sentences, test_qa_labels = prepare_sentences_and_spans(
        token_seq=test_token_seq,
        label_seq=test_label_seq,
    )

    test_dataset = Dataset(
        qa_sentences=test_qa_sentences,
        qa_labels=test_qa_labels,
        prompt_mapper=prompt_mapper,
    )

    collator = Collator(
        tokenizer=tokenizer,
        tokenizer_kwargs=tokenizer_kwargs,
    )

    train_dataloader = torch.utils.data.DataLoader(
        dataset=train_dataset,
        batch_size=args.batch_size,
        shuffle=True,
        collate_fn=collator,
    )

    test_dataloader = torch.utils.data.DataLoader(
        dataset=test_dataset,
        batch_size=args.batch_size,
        shuffle=False,
        collate_fn=collator,
    )

    # train / eval
    # TODO: change to AdamW
    optimizer = torch.optim.Adam(
        model.parameters(),
        lr=args.learning_rate,
    )

    train(
        n_epochs=args.n_epochs,
        model=model,
        train_dataloader=train_dataloader,
        test_dataloader=test_dataloader,
        optimizer=optimizer,
        writer=writer,
        device=device,
    )

    model.save_pretrained(args.path_to_save_model)
    tokenizer.save_pretrained(args.path_to_save_model)

    return 0


if __name__ == "__main__":
    main()



from collections import defaultdict

import numpy as np
import torch
from torch.utils.tensorboard import SummaryWriter
from tqdm import tqdm
from transformers import AutoModelForQuestionAnswering

from qaner.data_utils import Span
from qaner.inference_utils import get_top_valid_spans
from qaner.metrics import compute_metrics


# TODO: add metrics calculation
def train(
    n_epochs: int,
    model: AutoModelForQuestionAnswering,
    train_dataloader: torch.utils.data.DataLoader,
    test_dataloader: torch.utils.data.DataLoader,
    optimizer: torch.optim.Optimizer,
    writer: SummaryWriter,
    device: torch.device,
) -> None:
    """
    Training loop.

    Args:
        n_epochs (int): number of epochs to train.
        model (AutoModelForQuestionAnswering): model.
        train_dataloader (torch.utils.data.DataLoader): train_dataloader.
        test_dataloader (torch.utils.data.DataLoader): test_dataloader.
        optimizer (torch.optim.Optimizer): optimizer.
        writer (SummaryWriter): tensorboard writer.
        device (torch.device): cpu or cuda.
    """

    for epoch in range(n_epochs):

        print(f"Epoch [{epoch+1} / {n_epochs}]\n")

        train_epoch(
            model=model,
            dataloader=train_dataloader,
            optimizer=optimizer,
            writer=writer,
            device=device,
            epoch=epoch,
        )
        evaluate_epoch(
            model=model,
            dataloader=test_dataloader,
            writer=writer,
            device=device,
            epoch=epoch,
        )


def train_epoch(
    model: AutoModelForQuestionAnswering,
    dataloader: torch.utils.data.DataLoader,
    optimizer: torch.optim.Optimizer,
    writer: SummaryWriter,
    device: torch.device,
    epoch: int,
) -> None:
    """
    One training cycle (loop).

    Args:
        model (AutoModelForQuestionAnswering): QA model.
        dataloader (torch.utils.data.DataLoader): dataloader.
        optimizer (torch.optim.Optimizer): optimizer.
        writer (SummaryWriter): tensorboard writer.
        device (torch.device): cpu or cuda.
        epoch (int): number of current epochs.
    """

    model.train()

    epoch_loss = []
    batch_metrics_list = defaultdict(list)

    for i, inputs in tqdm(
        enumerate(dataloader),
        total=len(dataloader),
        desc="loop over train batches",
    ):
        optimizer.zero_grad()

        instances_batch = inputs.pop("instances")

        context_list, question_list = [], []
        for instance in instances_batch:
            context_list.append(instance.context)
            question_list.append(instance.question)

        inputs = inputs.to(device)
        offset_mapping_batch = inputs.pop("offset_mapping")

        outputs = model(**inputs)
        loss = outputs.loss
        loss.backward()
        optimizer.step()

        epoch_loss.append(loss.item())
        writer.add_scalar(
            "batch loss / train", loss.item(), epoch * len(dataloader) + i
        )

        with torch.no_grad():
            model.eval()
            outputs_inference = model(**inputs)
            model.train()

        spans_pred_batch_top_1 = get_top_valid_spans(
            context_list=context_list,
            question_list=question_list,
            prompt_mapper=dataloader.dataset.prompt_mapper,
            inputs=inputs,
            outputs=outputs_inference,
            offset_mapping_batch=offset_mapping_batch,
            n_best_size=1,
            max_answer_length=100,  # TODO: remove hardcode
        )

        # TODO: maybe move into get_top_valid_spans
        for idx in range(len(spans_pred_batch_top_1)):
            if not spans_pred_batch_top_1[idx]:
                empty_span = Span(
                    token="",
                    label="O",  # TODO: maybe not "O" label
                    start_context_char_pos=0,
                    end_context_char_pos=0,
                )
                spans_pred_batch_top_1[idx] = [empty_span]

        spans_true_batch = [instance.answer for instance in instances_batch]

        batch_metrics = compute_metrics(
            spans_true_batch=spans_true_batch,
            spans_pred_batch_top_1=spans_pred_batch_top_1,
            prompt_mapper=dataloader.dataset.prompt_mapper,
        )

        for metric_name, metric_value in batch_metrics.items():
            batch_metrics_list[metric_name].append(metric_value)
            writer.add_scalar(
                f"batch {metric_name} / train",
                metric_value,
                epoch * len(dataloader) + i,
            )

    avg_loss = np.mean(epoch_loss)
    print(f"Train loss: {avg_loss}\n")
    writer.add_scalar("loss / train", avg_loss, epoch)

    for metric_name, metric_value_list in batch_metrics_list.items():
        metric_value = np.mean(metric_value_list)
        print(f"Train {metric_name}: {metric_value}\n")
        writer.add_scalar(f"{metric_name} / train", metric_value, epoch)


# TODO: remove train_epoch code duplicates
def evaluate_epoch(
    model: AutoModelForQuestionAnswering,
    dataloader: torch.utils.data.DataLoader,
    writer: SummaryWriter,
    device: torch.device,
    epoch: int,
) -> None:
    """
    One evaluation cycle (loop).

    Args:
        model (AutoModelForQuestionAnswering): model.
        dataloader (torch.utils.data.DataLoader): dataloader.
        writer (SummaryWriter): tensorboard writer.
        device (torch.device): cpu or cuda.
        epoch (int): number of current epochs.
    """

    model.eval()

    epoch_loss = []
    batch_metrics_list = defaultdict(list)

    with torch.no_grad():

        for i, inputs in tqdm(
            enumerate(dataloader),
            total=len(dataloader),
            desc="loop over test batches",
        ):

            instances_batch = inputs.pop("instances")

            context_list, question_list = [], []
            for instance in instances_batch:
                context_list.append(instance.context)
                question_list.append(instance.question)

            inputs = inputs.to(device)
            offset_mapping_batch = inputs.pop("offset_mapping")

            outputs = model(**inputs)
            loss = outputs.loss

            epoch_loss.append(loss.item())
            writer.add_scalar(
                "batch loss / test", loss.item(), epoch * len(dataloader) + i
            )

            spans_pred_batch_top_1 = get_top_valid_spans(
                context_list=context_list,
                question_list=question_list,
                prompt_mapper=dataloader.dataset.prompt_mapper,
                inputs=inputs,
                outputs=outputs,
                offset_mapping_batch=offset_mapping_batch,
                n_best_size=1,  # TODO: remove hardcode
                max_answer_length=100,  # TODO: remove hardcode
            )

            # TODO: maybe move into get_top_valid_spans
            for idx in range(len(spans_pred_batch_top_1)):
                if not spans_pred_batch_top_1[idx]:
                    empty_span = Span(
                        token="",
                        label="O",  # TODO: maybe not "O" label
                        start_context_char_pos=0,
                        end_context_char_pos=0,
                    )
                    spans_pred_batch_top_1[idx] = [empty_span]

            # TODO: change metrics calculation for inference time
            spans_true_batch = [instance.answer for instance in instances_batch]

            batch_metrics = compute_metrics(
                spans_true_batch=spans_true_batch,
                spans_pred_batch_top_1=spans_pred_batch_top_1,
                prompt_mapper=dataloader.dataset.prompt_mapper,
            )

            for metric_name, metric_value in batch_metrics.items():
                batch_metrics_list[metric_name].append(metric_value)
                writer.add_scalar(
                    f"batch {metric_name} / test",
                    metric_value,
                    epoch * len(dataloader) + i,
                )

        avg_loss = np.mean(epoch_loss)
        print(f"Test loss:  {avg_loss}\n")
        writer.add_scalar("loss / test", avg_loss, epoch)

        for metric_name, metric_value_list in batch_metrics_list.items():
            metric_value = np.mean(metric_value_list)
            print(f"Test {metric_name}: {metric_value}\n")
            writer.add_scalar(f"{metric_name} / test", np.mean(metric_value), epoch)



import random

import numpy as np
import torch


def set_global_seed(seed: int):
    """
    Set global seed for reproducibility.

    Args:
        seed (int): Seed.
    """

    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    torch.backends.cudnn.benchmark = False
    torch.backends.cudnn.deterministic = True



import json
from typing import Any, Dict

import torch
from transformers import AutoModelForQuestionAnswering, AutoTokenizer

from qaner.arg_parse import get_inference_args
from qaner.data_utils import Instance
from qaner.inference_utils import get_top_valid_spans
from qaner.utils import set_global_seed


# TODO: add batch inference
def predict(
    context: str,
    question: str,  # TODO: change with entity type
    prompt_mapper: Dict[str, str],
    model: AutoModelForQuestionAnswering,
    tokenizer: AutoTokenizer,
    tokenizer_kwargs: Dict[str, Any],
    n_best_size: int = 1,
    max_answer_length: int = 100,
) -> Instance:
    """
    Predict function.

    Args:
        context (str): context string.
        question (str): question string.
        prompt_mapper (Dict[str, str]): prompt mapper.
        model (AutoModelForQuestionAnswering): trained QaNER model.
        tokenizer (AutoTokenizer): tokenizer for trained QaNER model.
        tokenizer_kwargs (Dict[str, Any]): tokenizer kwargs.
        n_best_size (int): number of best QA answers to consider. Defaults to 1.
        max_answer_length (int): entity max length to eliminate very long entities. Defaults to 100.

    Returns:
        Instance: predicted instance.
    """

    inputs = tokenizer([question], [context], **tokenizer_kwargs).to(model.device)
    offset_mapping_batch = inputs.pop("offset_mapping")

    with torch.no_grad():
        outputs = model(**inputs)

    spans_pred_batch_top = get_top_valid_spans(
        context_list=[context],
        question_list=[question],
        prompt_mapper=prompt_mapper,
        inputs=inputs,
        outputs=outputs,
        offset_mapping_batch=offset_mapping_batch,
        n_best_size=n_best_size,
        max_answer_length=max_answer_length,
    )[0]

    # TODO: validate it
    spans_pred_batch_top = [span for span in spans_pred_batch_top if span]

    for predicted_answer_span in spans_pred_batch_top:
        start_pos = predicted_answer_span.start_context_char_pos
        end_pos = predicted_answer_span.end_context_char_pos
        assert predicted_answer_span.token == context[start_pos:end_pos]

    prediction = Instance(
        context=context,
        question=question,
        answer=spans_pred_batch_top,
    )

    return prediction


def main() -> int:
    """
    Main inference function.

    Returns:
        int: exit code.
    """

    # argparse
    args = get_inference_args()

    set_global_seed(args.seed)
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

    # bert model
    tokenizer = AutoTokenizer.from_pretrained(args.path_to_trained_model)
    model = (
        AutoModelForQuestionAnswering.from_pretrained(args.path_to_trained_model)
        .eval()
        .to(device)
    )

    # TODO: validate it
    tokenizer_kwargs = {
        "max_length": 512,
        "truncation": "only_second",
        "padding": True,
        "return_tensors": "pt",
        "return_offsets_mapping": True,
    }

    with open(args.path_to_prompt_mapper, mode="r", encoding="utf-8") as fp:
        prompt_mapper = json.load(fp)

    prediction = predict(
        context=args.context,
        question=args.question,
        prompt_mapper=prompt_mapper,
        model=model,
        tokenizer=tokenizer,
        tokenizer_kwargs=tokenizer_kwargs,
        n_best_size=args.n_best_size,
        max_answer_length=args.max_answer_length,
    )

    print(f"\nquestion: {prediction.question}\n")
    print(f"context: {prediction.context}")
    print(f"\nanswer: {prediction.answer}\n")

    return 0


if __name__ == "__main__":
    main()



from collections import namedtuple
from typing import List, Tuple

from tqdm import tqdm, trange

Span = namedtuple(
    "Span", ["token", "label", "start_context_char_pos", "end_context_char_pos"]
)
Instance = namedtuple("Instance", ["context", "question", "answer"])


def read_bio_markup(
    path: str,
) -> Tuple[List[List[str]], List[List[str]]]:
    """
    Read BIO-markup data.
    Labels and tokens separated on each line with tab.
    Sentences are separated by empty line.

    Data example:
    label_11    token_11
    label_12    token_12
    label_21    token_21
    label_22    token_22
    label_23    token_23
    ...

    Args:
        path (str): Path to BIO-markup data.

    Returns:
        Tuple[List[List[str]], List[List[str]]]: Tokens and labels lists.
    """

    token_seq = []
    label_seq = []
    with open(path, mode="r") as fp:
        tokens = []
        labels = []
        for line in tqdm(fp, desc="read bio-markup"):
            if line != "\n":
                label, token = line.strip().split("\t")
                tokens.append(token)
                labels.append(label)
            else:
                if len(tokens) > 0:
                    token_seq.append(tokens)
                    label_seq.append(labels)
                tokens = []
                labels = []

    return token_seq, label_seq


def prepare_sentences_and_spans(
    token_seq: List[List[str]],
    label_seq: List[List[str]],
) -> Tuple[List[str], List[List[Span]]]:
    """
    Prepare sentences and label spans.

    Args:
        token_seq (List[List[str]]): Tokens lists.
        label_seq (List[List[str]]): Labels lists.

    Raises:
        Exception: Validate BIO-markup.

    Returns:
        Tuple[List[str], List[List[Span]]]: Sentences and label spans.
    """

    qa_sentences, qa_labels = [], []

    for i in trange(len(token_seq)):
        current_char_pos = 0
        qa_sent, qa_lab = [], []
        for token, label in zip(token_seq[i], label_seq[i]):
            qa_sent.append(token)
            if label != "O":
                span = Span(
                    token=token,
                    label=label,
                    start_context_char_pos=current_char_pos,
                    end_context_char_pos=current_char_pos + len(token),
                )
                qa_lab.append(span)
            current_char_pos += len(token) + 1
        qa_sentences.append(" ".join(qa_sent))
        qa_labels.append(qa_lab)

    qa_labels_v2 = []
    for qa_lab in qa_labels:
        qa_lab_v2 = []
        for span in qa_lab:
            if span.label.startswith("B-"):
                span_v2 = Span(
                    token=span.token,
                    label=span.label.split("-")[-1],
                    start_context_char_pos=span.start_context_char_pos,
                    end_context_char_pos=span.end_context_char_pos,
                )
                qa_lab_v2.append(span_v2)
            elif span.label.startswith("I-"):
                # TODO: remove duplicates and optimize
                span_v2 = Span(  # TODO: maybe use Span as dataclass
                    token=f"{span_v2.token} {span.token}",
                    label=span_v2.label,
                    start_context_char_pos=span_v2.start_context_char_pos,
                    end_context_char_pos=span.end_context_char_pos,
                )
                qa_lab_v2[-1] = span_v2
            else:
                raise ValueError(f"Unknown label: {span.label}")
        qa_labels_v2.append(qa_lab_v2)

    return qa_sentences, qa_labels_v2



from typing import Dict, List

import numpy as np

from qaner.data_utils import Span


# TODO: add metrics over label types
def compute_metrics(
    spans_true_batch: List[Span],
    spans_pred_batch_top_1: List[List[Span]],
    prompt_mapper: Dict[str, str],
) -> Dict[str, float]:
    """
    Compute NER metrics.

    Args:
        spans_true_batch (List[Span]): targets.
        spans_pred_batch_top_1 (np.ndarray): predictions.
        prompt_mapper (Dict[str, str]): prompt mapper.

    Returns:
        Dict[str, float]: metrics.
    """

    metrics = {}

    entity_mapper = {"O": 0}
    for entity_tag in prompt_mapper:
        entity_mapper[entity_tag] = len(entity_mapper)

    ner_confusion_matrix = np.zeros((len(entity_mapper), len(entity_mapper)))
    confusion_matrix_true_denominator = np.zeros(len(entity_mapper))
    confusion_matrix_pred_denominator = np.zeros(len(entity_mapper))

    for span_true, span_pred in zip(spans_true_batch, spans_pred_batch_top_1):
        span_pred = span_pred[0]  # type: ignore

        i = entity_mapper[span_true.label]
        j = entity_mapper[span_pred.label]  # type: ignore

        confusion_matrix_true_denominator[i] += 1
        confusion_matrix_pred_denominator[j] += 1

        if span_true == span_pred:
            ner_confusion_matrix[i, j] += 1

    assert (
        confusion_matrix_true_denominator.sum()
        == confusion_matrix_pred_denominator.sum()
    )

    ner_confusion_matrix_diag = np.diag(ner_confusion_matrix)

    # TODO: hide RuntimeWarning
    accuracy = np.nan_to_num(
        ner_confusion_matrix_diag.sum() / confusion_matrix_true_denominator.sum()
    )
    precision_per_entity_type = np.nan_to_num(
        ner_confusion_matrix_diag / confusion_matrix_pred_denominator
    )
    recall_per_entity_type = np.nan_to_num(
        ner_confusion_matrix_diag / confusion_matrix_true_denominator
    )
    f1_per_entity_type = np.nan_to_num(
        2
        * precision_per_entity_type
        * recall_per_entity_type
        / (precision_per_entity_type + recall_per_entity_type)
    )

    metrics["accuracy"] = accuracy

    for label_tag, idx in entity_mapper.items():
        metrics[f"precision_tag_{label_tag}"] = precision_per_entity_type[idx]
        metrics[f"recall_tag_{label_tag}"] = recall_per_entity_type[idx]
        metrics[f"f1_tag_{label_tag}"] = f1_per_entity_type[idx]

    # TODO: add micro average

    # macro average
    metrics["precision_macro"] = precision_per_entity_type.mean()
    metrics["recall_macro"] = recall_per_entity_type.mean()
    metrics["f1_macro"] = f1_per_entity_type.mean()

    # weighted average
    metrics["precision_weighted"] = np.average(
        precision_per_entity_type,
        weights=confusion_matrix_true_denominator,
    )
    metrics["recall_weighted"] = np.average(
        recall_per_entity_type,
        weights=confusion_matrix_true_denominator,
    )
    metrics["f1_weighted"] = np.average(
        f1_per_entity_type,
        weights=confusion_matrix_true_denominator,
    )

    return metrics



from typing import Any, Dict, List, Tuple

import torch
import transformers
from tqdm import tqdm

from qaner.data_utils import Instance, Span


# TODO: add documentation
class Dataset(torch.utils.data.Dataset):
    def __init__(
        self,
        qa_sentences: List[str],
        qa_labels: List[List[Span]],
        prompt_mapper: Dict[str, str],
    ):
        super().__init__()
        self.prompt_mapper = prompt_mapper
        self.dataset = self._prepare_dataset(
            qa_sentences=qa_sentences,
            qa_labels=qa_labels,
        )

    def __len__(self) -> int:
        return len(self.dataset)

    def __getitem__(self, idx: int) -> Instance:
        return self.dataset[idx]

    def _prepare_dataset(
        self,
        qa_sentences: List[str],
        qa_labels: List[List[Span]],
    ) -> List[Instance]:

        dataset = []

        for sentence, labels in tqdm(
            zip(qa_sentences, qa_labels),
            desc="prepare_dataset",
        ):
            for label_tag, label_name in self.prompt_mapper.items():
                question_prompt = f"What is the {label_name}?"

                answer_list = []
                for span in labels:
                    if span.label == label_tag:
                        answer_list.append(span)

                if len(answer_list) == 0:
                    empty_span = Span(
                        token="",
                        label="O",
                        start_context_char_pos=0,
                        end_context_char_pos=0,
                    )
                    instance = Instance(
                        context=sentence,
                        question=question_prompt,
                        answer=empty_span,
                    )
                    dataset.append(instance)
                else:
                    for answer in answer_list:
                        instance = Instance(
                            context=sentence,
                            question=question_prompt,
                            answer=answer,
                        )
                        dataset.append(instance)

        return dataset


# TODO: add documentation
class Collator:
    def __init__(
        self,
        tokenizer: transformers.PreTrainedTokenizer,
        tokenizer_kwargs: Dict[str, Any],
    ):
        self.tokenizer = tokenizer
        self.tokenizer_kwargs = tokenizer_kwargs

    def __call__(
        self,
        batch: List[Instance],
    ) -> transformers.tokenization_utils_base.BatchEncoding:

        context_list = []
        question_list = []
        start_end_context_char_pos_list = []

        for instance in batch:
            context_list.append(instance.context)
            question_list.append(instance.question)
            start_end_context_char_pos_list.append(
                [
                    instance.answer.start_context_char_pos,
                    instance.answer.end_context_char_pos,
                ]
            )

        tokenized_batch = self.tokenizer(
            question_list, context_list, **self.tokenizer_kwargs
        )

        offset_mapping_batch = tokenized_batch["offset_mapping"].numpy().tolist()

        assert len(offset_mapping_batch) == len(start_end_context_char_pos_list)

        # convert char boudaries to token boudaries in batch
        start_token_pos_list, end_token_pos_list = [], []
        for offset_mapping, start_end_context_char_pos in zip(
            offset_mapping_batch, start_end_context_char_pos_list
        ):
            # TODO: add comment
            if start_end_context_char_pos == [0, 0]:
                start_token_pos_list.append(0)
                end_token_pos_list.append(0)
            else:
                (
                    start_token_pos,
                    end_token_pos,
                ) = self.convert_char_boudaries_to_token_boudaries(
                    offset_mapping=offset_mapping,
                    start_end_context_char_pos=start_end_context_char_pos,
                )
                start_token_pos_list.append(start_token_pos)
                end_token_pos_list.append(end_token_pos)

        tokenized_batch["start_positions"] = torch.LongTensor(start_token_pos_list)
        tokenized_batch["end_positions"] = torch.LongTensor(end_token_pos_list)

        # additional
        tokenized_batch["instances"] = batch

        return tokenized_batch

    # TODO: add tests
    @staticmethod
    def convert_char_boudaries_to_token_boudaries(
        offset_mapping: List[List[int]],
        start_end_context_char_pos: List[int],
    ) -> Tuple[int, int]:

        start_context_char_pos, end_context_char_pos = start_end_context_char_pos
        assert end_context_char_pos >= start_context_char_pos

        done = False
        special_tokens_cnt = 0
        for i, token_boudaries in enumerate(offset_mapping):

            if token_boudaries == [0, 0]:
                special_tokens_cnt += 1
                continue

            if special_tokens_cnt == 2:
                start_token_pos, end_token_pos = token_boudaries

                if start_token_pos == start_context_char_pos:
                    res_start_token_pos = i

                if end_token_pos == end_context_char_pos:
                    res_end_token_pos = i  # inclusive
                    done = True
                    break

        assert done

        return res_start_token_pos, res_end_token_pos



from argparse import ArgumentParser, Namespace


def get_train_args() -> Namespace:
    """
    Training Argument Parser.

    Returns:
        Namespace: Parsed arguments.
    """

    parser = ArgumentParser()

    parser.add_argument(
        "--bert_model_name",
        type=str,
        required=True,
        help="bert model name",
    )
    parser.add_argument(
        "--path_to_prompt_mapper",
        type=str,
        required=True,
        help="path to prompt mapper",
    )
    parser.add_argument(
        "--path_to_train_data",
        type=str,
        required=True,
        help="path to train data",
    )
    parser.add_argument(
        "--path_to_test_data",
        type=str,
        required=True,
        help="path to test data",
    )
    parser.add_argument(
        "--path_to_save_model",
        type=str,
        required=True,
        help="path to save model",
    )
    parser.add_argument(
        "--n_epochs",
        type=int,
        required=True,
        help="number of epochs",
    )
    parser.add_argument(
        "--batch_size",
        type=int,
        required=True,
        help="batch size",
    )
    parser.add_argument(
        "--learning_rate",
        type=float,
        required=True,
        help="learning rate",
    )
    parser.add_argument(
        "--seed",
        type=int,
        required=False,
        default=42,
        help="seed for reproducibility",
    )
    parser.add_argument(
        "--log_dir",
        type=str,
        required=False,
        default="runs/qaner",
        help="tensorboard log_dir",
    )

    args = parser.parse_args()
    return args


def get_inference_args() -> Namespace:
    """
    Infrerence Argument Parser.

    Returns:
        Namespace: Parsed arguments.
    """

    parser = ArgumentParser()

    parser.add_argument(
        "--context",
        type=str,
        required=True,
        help="sentence to extract entities from",
    )
    parser.add_argument(
        "--question",
        type=str,
        required=True,
        help="question prompt with entity name to extract",
    )
    parser.add_argument(
        "--path_to_prompt_mapper",
        type=str,
        required=True,
        help="path to prompt mapper",
    )
    parser.add_argument(
        "--path_to_trained_model",
        type=str,
        required=True,
        help="path to trained QaNER model",
    )
    parser.add_argument(
        "--n_best_size",
        type=int,
        required=True,
        help="number of best QA answers to consider",
    )
    parser.add_argument(
        "--max_answer_length",
        type=int,
        required=False,
        default=100,
        help="entity max length",
    )
    parser.add_argument(
        "--seed",
        type=int,
        required=False,
        default=42,
        help="seed for reproducibility",
    )

    args = parser.parse_args()
    return args



from typing import Dict, List

import numpy as np
import torch
import transformers

from qaner.data_utils import Span


def get_top_valid_spans(
    context_list: List[str],
    question_list: List[str],
    prompt_mapper: Dict[str, str],
    inputs: transformers.tokenization_utils_base.BatchEncoding,
    outputs: transformers.tokenization_utils_base.BatchEncoding,
    offset_mapping_batch: torch.Tensor,
    n_best_size: int = 1,
    max_answer_length: int = 100,
) -> List[List[Span]]:
    """
    Inference function to get top valid spans.
    https://huggingface.co/docs/transformers/tasks/question_answering

    Args:
        context_list (List[str]): context strings in  batch.
        question_list (List[str]): question strings in batch.
        prompt_mapper (Dict[str, str]): prompt mapper.
        inputs (transformers.tokenization_utils_base.BatchEncoding): inputs.
        outputs (transformers.tokenization_utils_base.BatchEncoding): outputs.
        offset_mapping_batch (torch.Tensor): offset mapping.
        n_best_size (int): number of best spans. Defaults to 1.
        max_answer_length (int): max answer length. Defaults to 100.

    Returns:
        List[List[Span]]: top valid spans for each isntance in batch.
    """

    batch_size = len(offset_mapping_batch)

    inv_prompt_mapper = {v: k for k, v in prompt_mapper.items()}

    assert batch_size == len(context_list)
    assert batch_size == len(question_list)
    assert batch_size == len(inputs["input_ids"])
    assert batch_size == len(inputs["token_type_ids"])
    assert batch_size == len(outputs["start_logits"])
    assert batch_size == len(outputs["end_logits"])

    top_valid_spans_batch = []

    # TODO: optimize it
    for i in range(batch_size):
        context = context_list[i]

        offset_mapping = offset_mapping_batch[i].cpu().numpy()
        mask = inputs["token_type_ids"][i].bool().cpu().numpy()
        offset_mapping[~mask] = [0, 0]
        offset_mapping = [
            (span if span != [0, 0] else None) for span in offset_mapping.tolist()
        ]

        start_logits = outputs["start_logits"][i].cpu().numpy()
        end_logits = outputs["end_logits"][i].cpu().numpy()

        start_indexes = np.argsort(start_logits)[-1 : -n_best_size - 1 : -1].tolist()
        end_indexes = np.argsort(end_logits)[-1 : -n_best_size - 1 : -1].tolist()
        top_valid_spans = []

        # TODO: validate loop
        for start_index, end_index in zip(start_indexes, end_indexes):
            if (
                start_index >= len(offset_mapping)
                or end_index >= len(offset_mapping)
                or offset_mapping[start_index] is None
                or offset_mapping[end_index] is None
            ):
                continue
            if (end_index < start_index) or (
                end_index - start_index + 1 > max_answer_length
            ):
                continue
            if start_index <= end_index:
                start_context_char_char, end_context_char_char = offset_mapping[
                    start_index
                ]
                span = Span(
                    token=context[start_context_char_char:end_context_char_char],
                    label=inv_prompt_mapper[  # TODO: add inference exception
                        question_list[i].split(r"What is the ")[-1].rstrip(r"?")
                    ],
                    start_context_char_pos=start_context_char_char,
                    end_context_char_pos=end_context_char_char,
                )
                top_valid_spans.append(span)
        top_valid_spans_batch.append(top_valid_spans)
    return top_valid_spans_batch



from setuptools import setup

from qaner import __version__

with open("README.md", mode="r", encoding="utf-8") as fp:
    long_description = fp.read()


setup(
    name="qaner",
    version=__version__,
    description="Unofficial implementation of QaNER: Prompting Question Answering Models for Few-shot Named Entity Recognition.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Dani El-Ayyass",
    author_email="dayyass@yandex.ru",
    license_files=["LICENSE"],
    url="https://github.com/dayyass/qaner",
    packages=["qaner"],
    entry_points={
        "console_scripts": [
            "qaner-train = qaner.train:main",
            "qaner-inference = qaner.inference:main",
        ],
    },
    install_requires=[
        "numpy==1.21.6",
        "tensorboard==2.9.0",
        "torch==1.8.1",
        "tqdm==4.64.0",
        "transformers==4.19.2",
    ],
)
